#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

"""Advanced functional programming in Python.

.. note:: This module is in **EXPERIMENTAL** state, we encourage not to use it
   before declared stable.

"""
